// This program demonstrates the knapsack problem

#include <iostream>
#include "console.h"
#include "vector.h"
#include "simpio.h"

using namespace std;

struct WeightedObject {
    int weight;
    int value;
};

// function prototypes
int fillKnapsack(const Vector<WeightedObject> &objects, int targetWeight);
int fillKnapsackHelper(Vector<WeightedObject> &objects, int targetWeight);
void runTest(const Vector<int> &values, const Vector<int> &weights, int targetWeight, int solution);

int main() {
    runTest({12,10,8,11,14,7,9}, {4,6,5,7,3,1,6}, 18, 44);
    runTest({5,20,3,50,5,4,15,12,6,7}, {6,15,11,12,6,11,13,7,17,13}, 25, 67);
    runTest({3,4,5,6}, {2,3,4,5}, 5, 7);

    return 0;
}

void runTest(const Vector<int>& values, const Vector<int>& weights, int targetWeight, int solution) {
    Vector<WeightedObject> testObjects;

    for (int i=0; i < values.size(); i++) {
        WeightedObject object;
        object.value = values[i];
        object.weight = weights[i];
        testObjects.add(object);
    }

    cout << "Testing " << endl;
    cout << "values=" << values << endl;
    cout << "weights=" << weights << endl;
    cout << "with target weight " << targetWeight << " (should be " << solution << ")..." << endl;
    int answer = fillKnapsack(testObjects, targetWeight);
    if (answer == solution) {
        cout << "[CORRECT]" << endl;
    } else {
        cout << "[INCORRECT (response=" << answer << ")]" << endl;
    }

    cout << "----" << endl;
}

// Returns the max value of all items at most the given weight
int fillKnapsack(const Vector<WeightedObject>& objects,
                 int targetWeight) {
    Vector<WeightedObject> objectsCopy = objects;
    return fillKnapsackHelper(objectsCopy, targetWeight);
}


// Returns the max value of all items at most the given weight
int fillKnapsackHelper(Vector<WeightedObject> &objects, 
                       int targetWeight) {
    if (targetWeight == 0 || objects.isEmpty()) return 0;

    WeightedObject originalObject = objects.removeBack();

    // Attempt 1: try excluding this from the bag
    int best = fillKnapsackHelper(objects, targetWeight);

    // Attempt 2: try including this in the bag, if it fits
    if (originalObject.weight <= targetWeight) {
        int with = originalObject.value + 
                   fillKnapsackHelper(objects,
                   targetWeight - originalObject.weight);
        best = max(best, with);
    }

    objects.add(originalObject);
    return best;
}

